@@ -16,7 +16,7 @@
  *
  *  Implementation of the module '_rpy' and the 'Robj' type.
  *
- *  $Id: rpymodule.c,v 1.11 2004/02/17 18:48:27 warnes Exp $
+ *  $Id: rpymodule.c,v 1.20 2004/10/23 01:55:53 warnes Exp $
  *
  */
 
@@ -25,14 +25,9 @@
 
 #define NONAMELESSUNION
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
-#ifdef _WIN32
-#include <windows.h>
-#include <Startup.h>
-#endif
-
-
 /* Global objects */
 static SEXP get_item;
 static SEXP set_item;
@@ -157,10 +152,10 @@ seq_to_R(PyObject *obj)
   SEXP robj, rit;
   int i, len, state;
   int fsm[5][5] = {{0, 0, 0, 0, 0},
-		   {0, 1, 2, 3, 0},
-		   {0, 2, 2, 3, 0},
-		   {0, 3, 3, 3, 0},
-		   {0, 0, 0, 0, 4}};
+                   {0, 1, 2, 3, 0},
+                   {0, 2, 2, 3, 0},
+                   {0, 3, 3, 3, 0},
+                   {0, 0, 0, 0, 4}};
   
   len = PySequence_Length(obj);
   if (len == 0)
@@ -273,7 +268,7 @@ to_Rarray(PyObject *o)
 
   pytl = Py_BuildValue("[i]", tl);
   obj = (PyArrayObject *)PyArray_ContiguousFromObject((PyObject *)obj,
-						      PyArray_NOTYPE, 0, 0);
+                                                      PyArray_NOTYPE, 0, 0);
   nobj = PyArray_Reshape(obj, pytl);
   Py_XDECREF(pytl);
   Py_XDECREF(obj);
@@ -354,17 +349,17 @@ to_Robj(PyObject *obj)
 #endif
 
   } else if ((PySequence_Check(obj)) &&
-	     (PySequence_Size(obj) >= 0)) {
-    PROTECT(robj = seq_to_R(obj));	/* No labels */
+             (PySequence_Size(obj) >= 0)) {
+    PROTECT(robj = seq_to_R(obj));      /* No labels */
 
   } else if ((PyMapping_Check(obj)) &&
-	     (PyMapping_Size(obj) >= 0)) {
+             (PyMapping_Size(obj) >= 0)) {
     PROTECT(robj = dict_to_R(obj));
 
   } else {
     PyErr_Format(RPyExc_Exception, "cannot convert from type '%s'",
-		 obj->ob_type->tp_name);
-    PROTECT(robj = NULL);	/* Protected to avoid stack inbalance */
+                 obj->ob_type->tp_name);
+    PROTECT(robj = NULL);       /* Protected to avoid stack inbalance */
   }
 
   if (do_decref) {
@@ -402,7 +397,7 @@ to_PyDict(PyObject *obj, SEXP names)
  * fastest index */
 static PyObject *
 ltranspose(PyObject *list, int *dims, int *strides,
-	     int pos, int shift, int len)
+             int pos, int shift, int len)
 {
   PyObject *nl, *it;
   int i;
@@ -413,10 +408,10 @@ ltranspose(PyObject *list, int *dims, int *strides,
   if (pos == len-1) {
     for (i=0; i<dims[pos]; i++) {
       if (!(it = PyList_GetItem(list, i*strides[pos]+shift)))
-	return NULL;
+        return NULL;
       Py_INCREF(it);
       if (PyList_SetItem(nl, i, it) < 0)
-	return NULL;
+        return NULL;
     }
     return nl;
   }
@@ -511,12 +506,12 @@ to_Pyobj_basic(SEXP robj, PyObject **obj)
 #endif
 
   if (!robj)
-    return -1;			/* error */
+    return -1;                  /* error */
 
   if (robj == R_NilValue) {
     Py_INCREF(Py_None);
     *obj = Py_None;
-    return 1;			/* succeed */
+    return 1;                   /* succeed */
   }
 
   len = GET_LENGTH(robj);
@@ -528,42 +523,42 @@ to_Pyobj_basic(SEXP robj, PyObject **obj)
     case INTSXP:
       integers = INTEGER(robj);
       if(isFactor(robj)) {
-	thislevel = CHAR(STRING_ELT(GET_LEVELS(robj), integers[i]-1));
-	if (!(it = PyString_FromString(thislevel)))
-	  return -1;
+        thislevel = CHAR(STRING_ELT(GET_LEVELS(robj), integers[i]-1));
+        if (!(it = PyString_FromString(thislevel)))
+          return -1;
       }
       else {
-	if (!(it = PyInt_FromLong(integers[i])))
-	  return -1;
+        if (!(it = PyInt_FromLong(integers[i])))
+          return -1;
       }
       break;
     case REALSXP:
       reals = REAL(robj);
       if (!(it = PyFloat_FromDouble(reals[i])))
-	return -1;
+        return -1;
       break;
     case CPLXSXP:
       complexes = COMPLEX(robj);
       if (!(it = PyComplex_FromDoubles(complexes[i].r,
-				       complexes[i].i)))
-	return -1;
+                                       complexes[i].i)))
+        return -1;
       break;
     case STRSXP:
       strings = CHAR(STRING_ELT(robj, i));
       if (!(it = PyString_FromString(strings)))
-	return -1;
+        return -1;
       break;
     case LISTSXP:
       if (!(it = to_Pyobj_with_mode(elt(robj, i), 1)))
-	return -1;
+        return -1;
       break;
     case VECSXP:
       if (!(it = to_Pyobj_with_mode(VECTOR_ELT(robj, i), 1)))
-	return -1;
+        return -1;
       break;
     default:
       Py_DECREF(tmp);
-      return 0;			/* failed */
+      return 0;                 /* failed */
     }
     
     if (PyList_SetItem(tmp, i, it) < 0)
@@ -574,8 +569,8 @@ to_Pyobj_basic(SEXP robj, PyObject **obj)
   if (dim != R_NilValue) {
 #ifdef WITH_NUMERIC
     array = to_PyNumericArray(tmp, dim);
-    if (array) {		/* If the conversion to Numeric succeed.. */
-      *obj = array;		/* we are done */
+    if (array) {                /* If the conversion to Numeric succeed.. */
+      *obj = array;             /* we are done */
       Py_DECREF(tmp);
       return 1;
     }
@@ -631,9 +626,9 @@ from_class_table(SEXP robj)
     if (!fun) {
       PyErr_Clear();
       for (i=0; i<GET_LENGTH(rclass); i++)
-	if ((fun = PyDict_GetItemString(class_table,
-					CHAR(STRING_ELT(rclass, i)))))
-	  break;
+        if ((fun = PyDict_GetItemString(class_table,
+                                        CHAR(STRING_ELT(rclass, i)))))
+          break;
     }
     else
       Py_INCREF(fun);
@@ -689,16 +684,16 @@ to_Pyobj_proc(SEXP robj, PyObject **obj)
 
   i = from_proc_table(robj, &fun);
   if (i < 0)
-    return -1;			/* an error occurred */
+    return -1;                  /* an error occurred */
 
   if (!fun)
-    return 0;			/* conversion failed */
+    return 0;                   /* conversion failed */
 
   tmp = (PyObject *)Robj_new(robj, TOP_MODE);
   *obj = PyObject_CallFunction(fun, "O", tmp);
   Py_DECREF(fun);
   Py_DECREF(tmp);
-  return 1;			/* conversion succeed */
+  return 1;                     /* conversion succeed */
 }
 
 /* Convert a Robj to a Python object via the class table (mode 2) */
@@ -711,13 +706,13 @@ to_Pyobj_class(SEXP robj, PyObject **obj)
   fun = from_class_table(robj);
   
   if (!fun)
-    return 0;			/* conversion failed */
+    return 0;                   /* conversion failed */
   
   tmp = (PyObject *)Robj_new(robj, TOP_MODE);
   *obj = PyObject_CallFunction(fun, "O", tmp);
   Py_DECREF(fun);
   Py_DECREF(tmp);
-  return 1;			/* conversion succeed */
+  return 1;                     /* conversion succeed */
 }
 
 PyObject *
@@ -890,38 +885,38 @@ make_argl(int largl, PyObject *argl, SEXP *e)
     if(!it) goto fail_arg;
     if( PySequence_Size(it) != 2 )
       {
-	Py_DECREF(it);
-	goto fail_arg;
+        Py_DECREF(it);
+        goto fail_arg;
       }
     nobj = PySequence_GetItem(it, 0);
 
     /* Name can be a string, None, or NULL, error otherwise. */
     if (PyString_Check(nobj))
       {
-	name = dotter(PyString_AsString(nobj));
-	Py_DECREF(nobj);
+        name = dotter(PyString_AsString(nobj));
+        Py_DECREF(nobj);
       }
     else if (nobj == Py_None)
       {
-	name = NULL;
-	Py_DECREF(nobj);
+        name = NULL;
+        Py_DECREF(nobj);
       }
     else if(nobj == NULL)
       {
-	name = NULL;
+        name = NULL;
       }
     else
       {
-	Py_DECREF(nobj);
-	goto fail_arg;
+        Py_DECREF(nobj);
+        goto fail_arg;
       }
 
     /* Value can be anything. */
     value = PySequence_GetItem(it, 1);
     if (!value) 
       {
-	PyMem_Free(name);
-	goto fail;
+        PyMem_Free(name);
+        goto fail;
       }
 
     rvalue =  to_Robj(value);
@@ -935,8 +930,8 @@ make_argl(int largl, PyObject *argl, SEXP *e)
     /* Add name (if present) */
     if (name && strlen(name)>0) 
       {
-	SET_TAG(*e, Rf_install(name));
-	PyMem_Free(name);
+        SET_TAG(*e, Rf_install(name));
+        PyMem_Free(name);
       }
 
     /* Move index to new end of call */
@@ -946,7 +941,7 @@ make_argl(int largl, PyObject *argl, SEXP *e)
 
  fail_arg:
    PyErr_SetString(PyExc_ValueError, 
-	        "Argument must be a sequence of (\"name\", value) pairs.\n");
+                "Argument must be a sequence of (\"name\", value) pairs.\n");
  fail:
    return 0;
 }
@@ -976,7 +971,7 @@ Robj_lcall(PyObject *self, PyObject *args)
   if(largs != 1 || !PySequence_Check(args) )
     {
       PyErr_SetString(PyExc_ValueError, 
-	        "Argument must be a sequence of (\"name\", value) pairs.\n");
+                "Argument must be a sequence of (\"name\", value) pairs.\n");
       return NULL;
     }
 
@@ -1033,7 +1028,7 @@ Robj_autoconvert(PyObject *self, PyObject *args, PyObject *kwds)
   static char *kwlist[] = {"val", 0};
 
   if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:autoconvert", kwlist,
-				  &conversion))
+                                  &conversion))
     return NULL;
 
   if (conversion > 3) {
@@ -1060,7 +1055,7 @@ Robj_as_py(PyObject *self, PyObject *args, PyObject *kwds)
   int conv=default_mode;
 
   if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:as_py", kwlist,
-				   &conv))
+                                   &conv))
     return NULL;
 
   if (conv <= -2 || conv > TOP_MODE) {
@@ -1080,7 +1075,7 @@ static PyMethodDef Robj_methods[] = {
   {"local_mode", (PyCFunction)Robj_autoconvert, METH_VARARGS|METH_KEYWORDS},
   {"as_py", (PyCFunction)Robj_as_py, METH_VARARGS|METH_KEYWORDS},
   {"lcall", (PyCFunction)Robj_lcall, METH_VARARGS},
-  {NULL, NULL}		/* sentinel */
+  {NULL, NULL}          /* sentinel */
 };
 
 /* Sequence protocol implementation */
@@ -1168,14 +1163,14 @@ Robj_item(PyObject *a, int i)
 
 /* We should implement sq_slice, sq_contains ... */
 static PySequenceMethods Robj_as_sequence = {
-  (inquiry)Robj_len,				/* sq_length */
-  0,				/* sq_concat */
-  0,				/* sq_repeat */
-  (intargfunc)Robj_item,				/* sq_item */
-  0,				/* sq_slice */
-  (intobjargproc)Robj_ass_item,				/* sq_ass_item */
-  0,				/* sq_ass_slice */
-  0,				/* sq_contains */
+  (inquiry)Robj_len,                            /* sq_length */
+  0,                            /* sq_concat */
+  0,                            /* sq_repeat */
+  (intargfunc)Robj_item,                                /* sq_item */
+  0,                            /* sq_slice */
+  (intobjargproc)Robj_ass_item,                         /* sq_ass_item */
+  0,                            /* sq_ass_slice */
+  0,                            /* sq_contains */
 };
 
 
@@ -1193,30 +1188,30 @@ Robj_getattr(RobjObject *self, char *name)
 statichere PyTypeObject Robj_Type = {
   /* The ob_type field must be initialized in the module init function
    * to be portable to Windows without using C++. */
-#if defined(PRE_2_2)
+#if defined(PRE_2_2) || defined(_WIN32)    // Matjaz
   PyObject_HEAD_INIT(NULL)
 #else
   PyObject_HEAD_INIT(&PyType_Type)
 #endif
-  0,			/*ob_size*/
-  "Robj",		/*tp_name*/
-  sizeof(RobjObject),	/*tp_basicsize*/
-  0,			/*tp_itemsize*/
+  0,                    /*ob_size*/
+  "Robj",               /*tp_name*/
+  sizeof(RobjObject),   /*tp_basicsize*/
+  0,                    /*tp_itemsize*/
   /* methods */
   (destructor)Robj_dealloc, /*tp_dealloc*/
-  0,			/*tp_print*/
+  0,                    /*tp_print*/
 #ifdef PRE_2_2
   (getattrfunc)Robj_getattr,
 #else
   0,
 #endif
   0,
-  0,			/*tp_compare*/
-  0,			/*tp_repr*/
-  0,			/*tp_as_number*/
-  &Robj_as_sequence,			/*tp_as_sequence*/
-  0,			/*tp_as_mapping*/
-  0,			/*tp_hash*/
+  0,                    /*tp_compare*/
+  0,                    /*tp_repr*/
+  0,                    /*tp_as_number*/
+  &Robj_as_sequence,                    /*tp_as_sequence*/
+  0,                    /*tp_as_mapping*/
+  0,                    /*tp_hash*/
   (ternaryfunc)Robj_call,                      /*tp_call*/
   0,                      /*tp_str*/
 #if defined(PRE_2_2) || defined(_WIN32)
@@ -1273,7 +1268,7 @@ get(PyObject *self, PyObject *args, PyObject *kwds)
 
   static char *kwlist[] = {"name", "autoconvert", 0};
   if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|i:get", kwlist,
-				   &obj_str, &conversion))
+                                   &obj_str, &conversion))
     return NULL;
 
   robj = get_from_name(obj_str);
@@ -1323,7 +1318,7 @@ r_events(PyObject *self, PyObject *args, PyObject *kwds)
 
   static char *kwlist[] = {"usec", 0};
   if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:r_events", 
-				   kwlist, &usec))
+                                   kwlist, &usec))
     return NULL;
 
   if (R_interact) {
@@ -1369,20 +1364,27 @@ start_events(void)
 }
 
 
+void
+r_cleanup(void)
+{
+  R_CleanUp(SA_NOSAVE, 1, 0);
+}
+
 
 /* List of functions defined in the module */
 static PyMethodDef rpy_methods[] = {
-  {"get", (PyCFunction)get, METH_VARARGS | METH_KEYWORDS},
-  {"set_mode", (PyCFunction)set_mode, METH_VARARGS},
-  {"get_mode", (PyCFunction)get_mode, METH_VARARGS},
-  {"set_output", (PyCFunction)set_output, METH_VARARGS},
-  {"set_input", (PyCFunction)set_input, METH_VARARGS},
+  {"get",           (PyCFunction)get,           METH_VARARGS | METH_KEYWORDS},
+  {"set_mode",      (PyCFunction)set_mode,      METH_VARARGS},
+  {"get_mode",      (PyCFunction)get_mode,      METH_VARARGS},
+  {"set_output",    (PyCFunction)set_output,    METH_VARARGS},
+  {"set_input",     (PyCFunction)set_input,     METH_VARARGS},
   {"set_showfiles", (PyCFunction)set_showfiles, METH_VARARGS},
-  {"get_output", (PyCFunction)get_output, METH_VARARGS},
-  {"get_input", (PyCFunction)get_input, METH_VARARGS},
+  {"get_output",    (PyCFunction)get_output,    METH_VARARGS},
+  {"get_input",     (PyCFunction)get_input,     METH_VARARGS},
   {"get_showfiles", (PyCFunction)get_showfiles, METH_VARARGS},
-  {"r_events", (PyCFunction)r_events, METH_VARARGS | METH_KEYWORDS},
-  {NULL, NULL}		/* sentinel */
+  {"r_events",      (PyCFunction)r_events,      METH_VARARGS | METH_KEYWORDS},
+  {"r_cleanup",     (PyCFunction)r_cleanup,     METH_NOARGS},
+  {NULL, NULL}          /* sentinel */
 };
 
 #ifdef _WIN32
@@ -1448,15 +1450,15 @@ init_embedded_win32( void ) {
       }
 
       if (rc == ERROR_SUCCESS) {
-	/* set R_HOME */
+        /* set R_HOME */
        char *buf = (char *) malloc( (strlen( RHome ) + 8) * sizeof( char ));
        strcpy( buf, "R_HOME=");
        strcat( buf, RHome );
        putenv( buf );
       }
       else {
-	/* not found, fall back to getRHOME() */
-	strcpy(RHome, getRHOME());
+        /* not found, fall back to getRHOME() */
+        strcpy(RHome, getRHOME());
       }
     }
   }
@@ -1479,22 +1481,29 @@ init_embedded_win32( void ) {
 
   Rp->home = RUser;
   Rp->CharacterMode = LinkDLL;
-  Rp->ReadConsole = (blah1) R_ReadConsole;
-  Rp->WriteConsole = (blah2) R_WriteConsole;
+    
+  Rp->ReadConsole = (blah1) RPy_ReadConsole;    // Matjaz
+  Rp->WriteConsole = (blah2) RPy_WriteConsole;  // Matjaz
+     
   Rp->CallBack = (blah3) RPyDoNothing;
   Rp->message = char_message;
   Rp->yesnocancel = char_yesnocancel;
   Rp->busy = RPyBusy;
-  Rp->R_Quiet = 1;
+  Rp->R_Quiet = TRUE;
 
   /* run as "interactive", so server won't be killed after an error */
   Rp->R_Slave = Rp->R_Verbose = 0;
-  Rp->R_Interactive = 1;
-  Rp->RestoreAction = 0; /* no restore */
-  Rp->SaveAction = 2;    /* no save */
+  Rp->R_Interactive = TRUE;
+  Rp->RestoreAction = SA_RESTORE; /* no restore */
+  Rp->SaveAction    = SA_NOSAVE;    /* no save */
+  
+#if R_VERSION < 0x20000   // pre-R-2.0.0
+
   Rp->CommandLineArgs = NULL;
   Rp->NumCommandLineArgs = 0;
-
+#else
+  R_set_command_line_arguments(0, NULL);
+#endif
   R_SetParams(Rp); /* so R_ShowMessage is set */
   R_SizeFromEnv(Rp);
 
@@ -1511,6 +1520,7 @@ init_embedded_win32( void ) {
 DL_EXPORT(void)
 init_rpy(void)
 {
+  static char RHome[] = R_HOME;
   PyObject *m, *d;
   char *defaultargv[] = {"rpy", "-q", "--vanilla"};
   PyOS_sighandler_t old_int;
@@ -1558,9 +1568,21 @@ init_rpy(void)
   init_embedded_win32();
 #else
   //     envvar    c-def
-  setenv("R_HOME", R_HOME, 1);
+  if ( strlen(RHome)>0 )
+    {
+      setenv("R_HOME", RHome, 1);
+    }
+  else
+    {
+      // TODO: put code here to Run R RHOME to get R_HOME value
+      fprintf(stderr, "Error setting R_HOME at build time.\n");
+      exit(-1);
+    } 
+
+
+
   Rf_initEmbeddedR( sizeof(defaultargv) / sizeof(defaultargv[0]),
-		    defaultargv);
+                    defaultargv);
 #endif
 
   /* Restore Python handlers */
@@ -1610,3 +1632,29 @@ init_rpy(void)
   //  PyObject_Print(r_lock, stderr, Py_PRINT_RAW);
   r_lock = NULL;
 }
+
+
+#ifndef _WIN32
+char *getRHOME(void)
+{
+  char *cmd = "R RHOME";
+  char *buf;
+  FILE *ptr;
+  int index;
+
+  buf = malloc(BUFSIZ);
+  
+  if ((ptr = popen(cmd, "r")) != NULL)
+    fgets(buf, BUFSIZ, ptr);
+  (void) pclose(ptr);
+
+  index = strlen(buf);
+  index--; 
+
+  // strip off trailing newline or CR
+  while( buf[index]=='\n' || buf[index]=='\r' )
+    buf[index--] = 0;
+  
+  return buf;
+}
+#endif
